<?php

declare(strict_types=1);

/*
 * Chill is a software for social workers
 *
 * For the full copyright and license information, please view
 * the LICENSE file that was distributed with this source code.
 */

namespace Chill\DocStoreBundle\Validator\Constraints;

use Chill\DocStoreBundle\AsyncUpload\Exception\BadCallToRemoteServer;
use Chill\DocStoreBundle\AsyncUpload\Exception\TempUrlRemoteServerException;
use Chill\DocStoreBundle\AsyncUpload\TempUrlGeneratorInterface;
use Chill\DocStoreBundle\Entity\StoredObject;
use Symfony\Component\Validator\Constraint;
use Symfony\Component\Validator\ConstraintValidator;
use Symfony\Component\Validator\Exception\UnexpectedTypeException;
use Symfony\Component\Validator\Exception\UnexpectedValueException;
use Symfony\Contracts\HttpClient\Exception\HttpExceptionInterface;
use Symfony\Contracts\HttpClient\Exception\TransportExceptionInterface;
use Symfony\Contracts\HttpClient\HttpClientInterface;

final class AsyncFileExistsValidator extends ConstraintValidator
{
    public function __construct(
        private readonly TempUrlGeneratorInterface $tempUrlGenerator,
        private readonly HttpClientInterface $client
    ) {}

    public function validate($value, Constraint $constraint): void
    {
        if ($value instanceof StoredObject) {
            $this->validateObject($value->getFilename(), $constraint);
        } elseif (is_string($value)) {
            $this->validateObject($value, $constraint);
        } else {
            throw new UnexpectedValueException($value, StoredObject::class.' or string');
        }
    }

    protected function validateObject(string $file, Constraint $constraint): void
    {
        if (!$constraint instanceof AsyncFileExists) {
            throw new UnexpectedTypeException($constraint, AsyncFileExists::class);
        }

        $urlHead = $this->tempUrlGenerator->generate(
            'HEAD',
            $file,
            30
        );

        try {
            $response = $this->client->request('HEAD', $urlHead->url);

            if (404 === $status = $response->getStatusCode()) {
                $this->context->buildViolation($constraint->message)
                    ->setParameter('{{ filename }}', $file)
                    ->addViolation();
            } elseif (500 <= $status) {
                throw new TempUrlRemoteServerException($response->getStatusCode());
            } elseif (400 <= $status) {
                throw new BadCallToRemoteServer($response->getContent(false), $response->getStatusCode());
            }
        } catch (HttpExceptionInterface $exception) {
            if (404 !== $exception->getResponse()->getStatusCode()) {
                throw $exception;
            }
        } catch (TransportExceptionInterface $e) {
            throw new TempUrlRemoteServerException(0, previous: $e);
        }
    }
}
